#include <std_include.hpp>
#include "loader/component_loader.hpp"
#include "game/dvars.hpp"

#include <utils/hook.hpp>

namespace fastfile {
namespace {
bool is_using_mods() {
  return (*dvars::fs_gameDirVar) && *(*dvars::fs_gameDirVar)->current.string;
}

void db_build_os_path_from_source(const char* zone_name, game::FF_DIR source,
                                  unsigned int size, char* filename) {
  char user_map[game::MAX_QPATH]{};
  switch (source) {
  case game::FFD_DEFAULT:
    (void)sprintf_s(filename, size, "%s\\%s%s.ff",
                    std::filesystem::current_path().string().c_str(),
                    game::Sys_GetMapZoneDir(zone_name), zone_name);
    break;
  case game::FFD_MOD_DIR:
    assert(is_using_mods());

    (void)sprintf_s(filename, size, "%s\\%s\\%s.ff",
                    std::filesystem::current_path().string().c_str(),
                    (*dvars::fs_gameDirVar)->current.string, zone_name);
    break;
  case game::FFD_USER_MAP:
    game::I_strncpyz(user_map, zone_name, sizeof(user_map));

    (void)sprintf_s(filename, size, "%s\\%s\\%s\\%s.ff",
                    std::filesystem::current_path().string().c_str(),
                    "usermaps", user_map, zone_name);
    break;
  default:
    assert(false && "inconceivable");
    break;
  }
}

bool db_mod_file_exists() {
  if (!*(*dvars::fs_gameDirVar)->current.string)
    return false;

  char filename[game::MAX_OSPATH]{};
  db_build_os_path_from_source("mod", game::FFD_MOD_DIR, sizeof(filename),
                               filename);

  if (auto zone_file = game::Sys_OpenFileReliable(filename);
      zone_file != INVALID_HANDLE_VALUE) {
    CloseHandle(zone_file);
    return true;
  }

  return false;
}

void db_load_x_assets_stub(game::XZoneInfo* zone_info, unsigned int zone_count,
                           int sync_mode) {
  std::vector<game::XZoneInfo> zones(zone_info, zone_info + zone_count);

  if (db_mod_file_exists()) {
    zones.emplace_back("mod", game::DB_ZONE_COMMON, 0);
  }

  game::DB_LoadXAssets(zones.data(), zones.size(), sync_mode);
}

game::Sys_File sys_create_file_stub(const char* dir, const char* filename) {
  auto result = game::Sys_CreateFile(dir, filename);

  if (result.handle != INVALID_HANDLE_VALUE) {
    return result;
  }

  if (!is_using_mods()) {
    return result;
  }

  // .ff extension was added previously
  if (!game::I_stricmp(filename, "mod.ff") && db_mod_file_exists()) {
    char file_path[game::MAX_OSPATH]{};
    db_build_os_path_from_source("mod", game::FFD_MOD_DIR, sizeof(file_path),
                                 file_path);

    result.handle = game::Sys_OpenFileReliable(file_path);
  }

  return result;
}
} // namespace

class component final : public component_interface {
public:
  static_assert(sizeof(game::Sys_File) == 4);

  void post_load() override {
    utils::hook(0x50B637, db_load_x_assets_stub, HOOK_CALL)
        .install() // hook*
        ->quick();

    utils::hook(0x582FA2, sys_create_file_stub, HOOK_CALL)
        .install() // hook*
        ->quick();
  }
};
} // namespace fastfile

REGISTER_COMPONENT(fastfile::component)
